unit Srv_Cli;

interface

uses WSocket, WSocketS, sysutils, Forms, Windows;

const
  SrvCliSignature : String = 'gyz01';

type
  TCommand=(
    cmdNone, cmdGo, cmdError, cmdLogin, cmdLogout,
    cmdTime, cmdSQL, cmdLoadFile, cmdWho
  );
  TSQLCommand=(cmdSQLExec, cmdSQLOpen, cmdSQLFieldDefs);
  TCommandName=record
    Command: TCommand;
    Name:    String;
  end;
  PInteger = ^Integer;
  PConnBuffer = ^TConnBuffer;
  TConnBuffer = class(TObject)
  protected
    FReadParamPos   : Integer;
    FTotalParamSize,
    FVirtualParamSize: Integer;
    FUnusedData     : String;
    FCommand        : TCommand;
    FCommandParams  : String;
    FResultParams   : String;
    FWaiting        : Boolean;
    FUpdating       : String;
    {$IFNDEF TESTING}
      FSocket: TWSocket;
    {$ENDIF}
    procedure ProcessData; virtual; abstract;
  public
    {$IFDEF TESTING}
      ReceiveStr, SendStr: String;
    {$ENDIF}

    LoginTime       : TDateTime;
    LoginName       : ShortString;
    property  Command: TCommand read FCommand;

    constructor Create( Socket: TWSocket ); virtual;
    procedure SetCommand( Cmd: TCommand; const Param: String='');
    function  ReadByte:Byte;
    function  ReadInt:Integer;
    function  ReadParam:String;
    procedure WriteByte( Param : Byte);
    procedure WriteInt( Param : Integer);
    procedure WriteParam(const Param:String);
    procedure Wait( Cmd: TCommand; MiliSeconds: Integer );
    procedure Parse;
    procedure Send;
    procedure Error( const Msg: String ); virtual;
    procedure Log( const Msg:String); virtual; abstract;
    procedure BeginUpdate( const Notes:String );
    procedure EndUpdate;
    {$IFDEF TESTING}
    function  Testing:String;
    {$ENDIF}
  end;

const
  CommandNames:array [low(TCommand)..High(TCommand)] of TCommandName=(
    (Command:cmdNone; Name:'None'), // Sep
    (Command:cmdGo; Name:'Go'),
    (Command:cmdError; Name:'Error'),
    (Command:cmdLogin; Name:'Login'),
    (Command:cmdLogout; Name:'Logout'),
    (Command:cmdTime; Name:'Time'),
    (Command:cmdSQL; Name:'SQL'),
    (Command:cmdLoadFile; Name:'LoadFile'),
    (Command:cmdWho; Name:'Who')
  );

function  strDivide(var txt:string; const Separador:string):string;


implementation

constructor TConnBuffer.Create( Socket: TWSocket );
begin
  Inherited Create;
  {$IFNDEF TESTING}
  FSocket := Socket;
  {$ENDIF}
  FCommand := cmdNone;
  FTotalParamSize := 0;
  FVirtualParamSize:= 0;
  LoginTime := Now;
end;

procedure TConnBuffer.SetCommand( Cmd: TCommand; const Param: String='');
begin
  FCommand := Cmd;
  FReadParamPos := 1;
  FResultParams  := Char(Ord( FCommand )) + SrvCliSignature + '1234' ;
  if Param<>'' then WriteParam( Param );
end;

function  TConnBuffer.ReadByte:Byte;
begin
  Result := 255;
  if FReadParamPos <= Length(FCommandParams) then begin
    Result := Byte(FCommandParams[FReadParamPos]);
    Inc(FReadParamPos);
  end;
end;

function  TConnBuffer.ReadInt:Integer;
begin
  Result := -1;
  if FReadParamPos <= Length(FCommandParams)-SizeOf(Result) then begin
    Result := PInteger( @FCommandParams[FReadParamPos] )^;
    Inc( FReadParamPos, SizeOf( Result ));
  end;
end;

function  TConnBuffer.ReadParam:String;
var
  Size: Integer;
begin
  if FReadParamPos > Length(FCommandParams) then Result := ''
  else begin
    Size := PInteger( @FCommandParams[FReadParamPos] )^;
    Inc( FReadParamPos, SizeOf( Size ));
    SetLength(Result, Size);
    if Size>0 then
      move (FCommandParams[FReadParamPos], Result[1], Size);
    Inc( FReadParamPos, Size);
  end;
end;

procedure TConnBuffer.WriteByte( Param : Byte);
begin
  FResultParams := FResultParams + Char(Param);
end;

procedure TConnBuffer.WriteInt( Param : Integer);
var
  SizeR : Integer;
begin
  SizeR := Length( FResultParams );
  SetLength(FResultParams, SizeR + SizeOf(Integer));
  PInteger( @FResultParams[SizeR+1] )^ := Param;
end;

procedure TConnBuffer.WriteParam(const Param:String);
var
  SizeR, SizeP : Integer;
begin
  SizeR := Length( FResultParams );
  SizeP := Length( Param );
  SetLength(FResultParams, SizeR + SizeOf(Integer) + SizeP);
  PInteger(@FResultParams[SizeR+1])^ := SizeP;
  if SizeP>0 then 
    move (Param[1],FResultParams[SizeR + SizeOf(Integer)+1], SizeP);
end;

procedure TConnBuffer.Wait;
var
  FormEnableds: array of boolean;
  I: Integer;
  Ini, Atu: TTimeStamp;
  Dif: Integer;
begin
  SetLength(FormEnableds, Screen.CustomFormCount);

  for I:=low(FormEnableds) to low(FormEnableds) do begin
    FormEnableds[I] := Screen.CustomForms[I].Enabled;
    Screen.CustomForms[I].Enabled := False;
  end;

  try

    FWaiting := True;
    FTotalParamSize := 0;
    FVirtualParamSize:= 0;

    {$IFNDEF TESTING}
    Ini := DateTimeToTimeStamp(Now);

    while FWaiting do begin
      FSocket.ProcessMessages;
      Sleep(0);
      Atu := DateTimeToTimeStamp(Now);
      Dif := Atu.Time - Ini.Time;
      if (Dif > Miliseconds) or (Dif<0) then begin
        FWaiting := False;
        Error( 'Wait timeout');
      end;
    end;
    {$ENDIF}

  finally
    for I:=low(FormEnableds) to low(FormEnableds) do begin
      Screen.CustomForms[I].Enabled := FormEnableds[I];
    end;
  end;

  if FCommand <> Cmd then Error('O servidor retornou um comando diferente do esperado');

end;

procedure TConnBuffer.Parse;
var SizeB, UnusedSize, UsedSize, HeaderSize: Integer;
    S : String;
begin

  S := FUnusedData +
       {$IFNDEF TESTING} FSocket. {$ENDIF}
       ReceiveStr;
  SizeB := Length(S);

  {$B-}
  if (FVirtualParamSize>0) or // Existe um comando em aberto ou Assinatura est Ok
     ((SizeB>=(1+SizeOf(FTotalParamSize)+Length(SrvCliSignature)))
     and CompareMem(@S[2], @SrvCliSignature[1], Length(SrvCliSignature))) then begin

    FUnusedData := '';

    if SizeB >0 then begin

      if FVirtualParamSize=0 then begin  // inicio do comando

        SetCommand( TCommand(Ord(S[1])) );

        FTotalParamSize := PInteger( @S[ Length(SrvCliSignature) + 2] )^;
        HeaderSize := 1 + Length(SrvCliSignature) + SizeOf(FTotalParamSize);

        SetLength( FCommandParams, FTotalParamSize);

      end
      else begin
        HeaderSize := 0;
      end;

      UsedSize   := FTotalParamSize - FVirtualParamSize;
      if UsedSize> (SizeB - HeaderSize) then UsedSize := (SizeB - HeaderSize);
      UnusedSize := SizeB - UsedSize - HeaderSize;

      move( S[ HeaderSize+1 ], FCommandParams[ FVirtualParamSize +1 ], UsedSize );
      Inc ( FVirtualParamSize, UsedSize );

      if FVirtualParamSize>=FTotalParamSize then begin
        FVirtualParamSize := 0;
        ProcessData;
        FWaiting := False;
      end;

      if UnusedSize>0 then begin
        FUnusedData := Copy( S, HeaderSize + UsedSize + 1, UnusedSize )
      end;

    end;
  end
  else Log( S );
end;

procedure TConnBuffer.Send;
var
  Size: Integer;
  HeaderSize : Integer;
  QtdeBytes : Integer;
procedure SocketConnect;
begin
  FSocket.Connect;
  while FSocket.State in [wsConnecting] do FSocket.ProcessMessage;
  if FSocket.State <> wsConnected then
    Error('Impossivel conectar');
end;
begin

  {$IFNDEF TESTING}
  if FSocket.State <> wsConnected then begin
    SocketConnect;
  end;
  {$ENDIF}

  Size := Length(FResultParams);
  HeaderSize := 1 + Length(SrvCliSignature) ;
  PInteger( @FResultParams[ HeaderSize +1 ] )^ := Size -
    SizeOf(FTotalParamSize) - HeaderSize;

  {$IFDEF TESTING}
    SendStr := FResultParams;
  {$ELSE}
  try
    QtdeBytes := FSocket.Send(@FResultParams[1], Size);
  except
    SocketConnect;
    QtdeBytes := FSocket.Send(@FResultParams[1], Size);
  end;

  if QtdeBytes <> Size then
      Error('erro enviando comando');
  {$ENDIF}

end;

procedure TConnBuffer.Error( const Msg: String );
begin
  Log ( Msg );
end;

procedure TConnBuffer.BeginUpdate;
begin
  if FUpdating<>'' then Error(' BeginUpdate nao pode executar. '#13+ FUpdating + #13+ Notes);
  FUpdating := Notes;
end;

procedure TConnBuffer.EndUpdate;
begin
  FUpdating := '';
end;

{$IFDEF TESTING}
function TConnBuffer.Testing:String;
var s1,s2, s3:string;
    i:integer;
begin

  setLength(s1, 256);
  for i:=0 to 255 do
    s1[i+1] := char(i);

  setLength(s2, 25005);
  for i:=1 to 25005 do
    s2[i] := char( random(255) ); //  '123456789'#0'1234567890123456789'#255'1234'

  SetCommand( cmdGo );

  WriteParam( s1 );
  WriteParam( s2 );

  WriteParam( '1234567890123456789012345678901234' );
  WriteByte( 10 );
  WriteInt( -10 );
  WriteParam( 'wp2' );

  WriteParam( s1 );
  WriteParam( s2 );

  Send;
  ReceiveStr := SendStr;
  Parse;

  case Command of
    cmdGo: begin
      if s1<> ReadParam then Result := Result + 'erro em s1-a '#13;
      if s2<> ReadParam then Result := Result + 'erro em s2-a'#13;
      Result :=               ReadParam + #13;
      Result := Result + IntToStr( ReadByte ) +#13;
      Result := Result + IntToStr( ReadInt ) + #13;
      Result := Result +           ReadParam + #13;
      if s1<> ReadParam then Result := Result + 'erro em s1-a '#13;
      s3 := ReadParam;
      if s2<> s3 then Result := Result + 'erro em s2-b '#13;
    end;
  end;
end;
{$ENDIF}

function strDivide;
  var i:integer;
begin
  i:=pos(Separador,txt);
  if i<1 then begin
    Result:=txt;
    txt:='';
  end
  else begin
    Result:=copy(txt,1,i-1);
    txt:=copy(txt,i+length(Separador),length(txt)-i);
  end;
end;

end.


